/***********************************************************/
/* Camera Thread                                           */
/* The most important class of the application             */
/* This thread controls the capture process                */
/***********************************************************/

#include <qmetatype.h>
#include "CameraThread.h"

#include <FCam/N900.h>

#include <vector>
#include <list>
#include <iostream>

#include "OverlayWidget.h"
#include "arm_neon.h"

#define HDR_NORMAL 0
#define HDR_LOW 1
#define HDR_HIGH 2

#define BURSTSIZE 3
#define VIDEOSIZE 102
using namespace std;

namespace Plat = FCam::N900;

//performs the autoExposure without changing the gain
void CameraThread::autoExpose(FCam::Shot *s, const FCam::Frame &f,
                    int maxExposure,
                    float smoothness) {
        if (!s) return;

        int b = f.histogram().buckets();

        int cdf[256]; // a CDF for histograms of up to 256 buckets

        if (!f.histogram().valid()) return;
        if (b > 256 || b < 64) return;

        const FCam::Histogram &hist = f.histogram();

        cdf[0] = hist(0);
        for (int i = 1; i < b; i++) {
            cdf[i] = cdf[i-1] + hist(i);
        }

        int brightPixels = cdf[b-1] - cdf[b-21]; // top 20 buckets
        int targetBrightPixels = cdf[b-1]/50;
        int maxSaturatedPixels = cdf[b-1]/200;
        int saturatedPixels = cdf[b-1] - cdf[b-6]; // top 5 buckets

        // how much should I change brightness by
        float adjustment = 1.0f;

        if (saturatedPixels > maxSaturatedPixels) {
            // first don't let things saturate too much
            adjustment = 1.0f - ((float)(saturatedPixels - maxSaturatedPixels))/cdf[b-1];
        } else if (brightPixels < targetBrightPixels) {
            // increase brightness to try and hit the desired number of well exposed pixels
            int l = b-11;
            while (brightPixels < targetBrightPixels && l > 0) {
                brightPixels += cdf[l];
                brightPixels -= cdf[l-1];
                l--;
            }

            // that level is supposed to be at b-11;
            adjustment = float(b-11+1)/(l+1);
        } else {
            // we're not oversaturated, and we have enough bright pixels. Do nothing.
        }

        if (adjustment > 4.0) adjustment = 4.0;
        if (adjustment < 1/16.0f) adjustment = 1/16.0f;

        float brightness =  f.exposure();
        float desiredBrightness = brightness * adjustment;
        int exposure;
        // Apply the smoothness constraint
        float shotBrightness =  s->exposure;
        desiredBrightness = shotBrightness * smoothness + desiredBrightness * (1-smoothness);

        // whats the largest we can raise exposure without negatively
        // impacting frame-rate or introducing handshake. We use 1/30s
        int exposureKnee = 33333;

        if (desiredBrightness > exposureKnee) {
            exposure = exposureKnee;
        } else {
            exposure = desiredBrightness;
        }

        // Finally, clamp the exposure at max
        if (exposure > maxExposure) {
            exposure = maxExposure;
        }

        s->exposure  = exposure;
        s->gain      = f.gain();
    }



void CameraThread::run() {
    // Make an asynchronous file writer to save images in the background
    FCam::AsyncFileWriter writer;
    Plat::Sensor sensor;
    Plat::Lens lens;

    // tell the sensor that the lens will be tagging
    // frames that come back from it    ;
    sensor.attach(&lens);

    // Make a helper autofocus object
    FCam::AutoFocus autoFocus(&lens);

    // The viewfinder shot
    FCam::Shot viewfinder;
    viewfinder.exposure = 40000;
    viewfinder.gain = 1.0f;
    // run at 25 fps
    viewfinder.frameTime = 40000;
    // dump image data directly into the frame buffer
    viewfinder.image = overlay->framebuffer();
    // enable histograms and sharpness maps
    viewfinder.histogram.enabled = true;
    viewfinder.histogram.region = FCam::Rect(0, 0, 640, 480);
    viewfinder.sharpness.enabled = true;
    viewfinder.sharpness.size = FCam::Size(16, 12);


    bool takeSnapshot = false;
    bool halfDepress = false;
    bool fullDepress = false;
    bool capturing = false;

    // stream the viewfinder
    sensor.stream(viewfinder);

    float gain = 1.f;
    int aux = 0;

    std::vector<FCam::Image> img(VIDEOSIZE);
    std::vector<FCam::Shot> shot(VIDEOSIZE); //must be a multiple of 3

    //allocates all images
    for(int i = 0; i < VIDEOSIZE; ++i){
        img[i] = FCam::Image(overlay->framebuffer().size(), FCam::UYVY);
     }

    int nbursts = 0;
    while (keepGoing){
        // deal with FCam events
        FCam::Event e;
        while (FCam::getNextEvent(&e)) {
            cout << e.description << endl;
            switch(e.type) {
            case FCam::Event::FocusPressed:
                halfDepress = true;
                break;
            case FCam::Event::FocusReleased:
                halfDepress = false;
                break;
            case FCam::Event::ShutterPressed:
                takeSnapshot = true;
                fullDepress = true;

                if(!capturing){
                    capturing = true;
                    emit startCapturing();
                    sensor.stopStreaming();
                    while (sensor.shotsPending()) sensor.getFrame(); //consumes the streaming queue
                    gain = viewfinder.gain;
                    //setup of the capture parameters
                    for(int i = 0; i < BURSTSIZE; i+=3){
                        shot[i].exposure = viewfinder.exposure;
                        shot[i].gain = gain; //overlay->framebuffer_
                        img[i] = FCam::Image(overlay->framebuffer().size(), FCam::UYVY);//, FCam::Image::AutoAllocate);
                        shot[i].image = img[i];
                        shot[i].whiteBalance = viewfinder.whiteBalance;
                        shot[i].frameTime = 40000;
                        shot[i].histogram.enabled = true;
                        shot[i].histogram.region = FCam::Rect(0, 0, img[i].width(), img[i].height());

                        shot[i+1].exposure = shot[i].exposure * 3;
                        shot[i+1].gain = gain;
                        img[i+1] = FCam::Image(overlay->framebuffer().size(), FCam::UYVY);//, FCam::Image::AutoAllocate);
                        shot[i+1].image = img[i+1];
                        shot[i+1].frameTime = 40000;
                        shot[i+1].whiteBalance = viewfinder.whiteBalance;
                        shot[i+1].histogram.enabled = true;
                        shot[i+1].histogram.region = FCam::Rect(0, 0, img[i].width(), img[i].height());

                        shot[i+2].exposure = (int)shot[i].exposure * 1.f/3.f;
                        shot[i+2].gain = gain;
                        img[i+2] = FCam::Image(overlay->framebuffer().size(), FCam::UYVY);//, FCam::Image::AutoAllocate);
                        shot[i+2].image = img[i+2];
                        shot[i+2].frameTime = 40000;
                        shot[i+2].whiteBalance = viewfinder.whiteBalance;
                        shot[i+2].histogram.enabled = true;
                        shot[i+2].histogram.region = FCam::Rect(0, 0, img[i].width(), img[i].height());
                    }
                    //captures BURSTSIZE photos
                    std::vector<FCam::Shot>::const_iterator first = shot.begin() + nbursts*BURSTSIZE;
                    std::vector<FCam::Shot>::const_iterator last = shot.begin() + BURSTSIZE;
                    for (first; first < last; ++first){
                        sensor.capture(*first);
                    }
                    ++nbursts;
                }
                break;
            case FCam::Event::ShutterReleased:
                fullDepress = false;
            };
        }

        if(capturing){
            if(aux < VIDEOSIZE)
             {
                FCam::Frame f = sensor.getFrame();
              if(aux%BURSTSIZE == 0){
                    // Make burst of shots
                    //memcpy(overlay->ptr,f.image()(0,0),640*480*2);
                    for (size_t y = 0; y < overlay->framebuffer().height(); y ++) {
                      uint8x16_t *outPtr = (uint8x16_t *)overlay->framebuffer()(0,y);
                      uint8x16_t *loPtr  = (uint8x16_t *)f.image()(0,y);
                      for (size_t x = 0; x < overlay->framebuffer().width()/8; x++) {
                          *outPtr++ = *loPtr++;
                      }
                    }
                    FCam::Shot saux;
                    saux.exposure =  shot[(nbursts-1)*BURSTSIZE].exposure; //picking the previous parameters
                    saux.gain =  shot[(nbursts-1)*BURSTSIZE].gain;
                    saux.whiteBalance =  shot[(nbursts-1)*BURSTSIZE].whiteBalance;
                    this->autoExpose(&saux, f);

                    int i = 0;
                    //setting up another burst, autoexposed
                    for(i = nbursts*BURSTSIZE; (i < (nbursts+1)*BURSTSIZE) && (i < VIDEOSIZE); i+=3){
                        shot[i].exposure = saux.exposure;
                        shot[i].gain = viewfinder.gain;
                        shot[i].image = img[i];
                        shot[i].frameTime = 40000;
                        shot[i].whiteBalance = viewfinder.whiteBalance;
                        shot[i].histogram.enabled = true;
                        shot[i].histogram.region = FCam::Rect(0, 0, img[i].width(), img[i].height());

                        shot[i+1].exposure = saux.exposure * 3;
                        shot[i+1].gain = viewfinder.gain;
                        shot[i+1].image = img[i+1];//, FCam::Image::AutoAllocate);
                        shot[i+1].frameTime = 40000;
                        shot[i+1].whiteBalance = viewfinder.whiteBalance;
                        shot[i+1].histogram.enabled = true;
                        shot[i+1].histogram.region = FCam::Rect(0, 0, img[i].width(), img[i].height());

                        shot[i+2].exposure = (int)saux.exposure * 1.f/3.f;
                        shot[i+2].gain = viewfinder.gain;
                        shot[i+2].image = img[i+2];//, FCam::Image::AutoAllocate);
                        shot[i+2].frameTime = 40000;
                        shot[i+2].whiteBalance = viewfinder.whiteBalance;
                        shot[i+2].histogram.enabled = true;
                        shot[i+2].histogram.region = FCam::Rect(0, 0, img[i].width(), img[i].height());
                        }
                    std::vector<FCam::Shot>::const_iterator first = shot.begin() + nbursts*BURSTSIZE;
                    std::vector<FCam::Shot>::const_iterator last = shot.begin() + i;
                    for (first; first < last; ++first){
                        sensor.capture(*first);
                    }
                    ++nbursts;
                }
                ++aux;
             }
            else //shots pending == 0
                {
                sensor.stopStreaming();
                while (sensor.shotsPending()) sensor.getFrame(); //consumes the streaming queue

                //preparing the next save
                FILE* nsave = fopen("/home/user/MyDocs/nsave.conf","rt");
                int nsaves = 0;
                fscanf(nsave,"%d",&nsaves);
                fclose(nsave);

                //saving the current photos
                capturing = false;
                //keepGoing = false;
                emit finishCapturing();
                unsigned int j = 0;
                while(j < img.size())
                {
                        char fname[256];
                        snprintf(fname, 255,  "%s/foto%4.4d.jpg",path,j);
                        FCam::saveJPEG(img[j], fname,100);
                        ++j;
                        emit curSave((double)j*100/img.size());
                }

                char dirName[550];
                sprintf(dirName,"/home/user/MyDocs/capture%d/shift0",nsaves);
                mkdir(dirName, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
                for(unsigned int i = 0; i < shot.size()-3; i+=3){
                    char fname[256];
                    snprintf(fname, 255, "%s/trinca%4.4d.hdrgen",dirName,i);
                    FILE* hdrgen = fopen(fname,"wt");
                    char trinca[1024];
                    // Gain as a float from 1.0 to 32 (equivalent to ISO 100-3200)
                    sprintf(trinca, "foto%4.4d.jpg %f 2.8 %d 0\nfoto%4.4d.jpg %f 2.8 %d 0\nfoto%4.4d.jpg %f 2.8 %d 0",
                                    i,1.0/((float)shot[i].exposure /1000000.0),(int)(shot[i].gain*100.f + 0.5),
                                    i+1,1./((float)shot[i+1].exposure /1000000.),(int)(shot[i+1].gain*100.f + 0.5),
                                    i+2,1./((float)shot[i+2].exposure /1000000.), (int)(shot[i+2].gain*100.f + 0.5));
                    fprintf(hdrgen,trinca);
                    fclose(hdrgen);
                }

                sprintf(dirName,"/home/user/MyDocs/capture%d/shift1",nsaves);
                mkdir(dirName, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
                for(unsigned int i = 1; i < shot.size()-3; i+=3){
                    char fname[256];
                    snprintf(fname, 255, "%s/trinca%4.4d.hdrgen",dirName,i);
                    FILE* hdrgen = fopen(fname,"wt");
                    char trinca[1024];
                    // Gain as a float from 1.0 to 32 (equivalent to ISO 100-3200)
                    sprintf(trinca, "foto%4.4d.jpg %f 2.8 %d 0\nfoto%4.4d.jpg %f 2.8 %d 0\nfoto%4.4d.jpg %f 2.8 %d 0",
                                    i,1.0/((float)shot[i].exposure /1000000.0),(int)(shot[i].gain*100.f + 0.5),
                                    i+1,1./((float)shot[i+1].exposure /1000000.),(int)(shot[i+1].gain*100.f + 0.5),
                                    i+2,1./((float)shot[i+2].exposure /1000000.), (int)(shot[i+2].gain*100.f + 0.5) );
                    fprintf(hdrgen,trinca);
                    fclose(hdrgen);
                }

                sprintf(dirName,"/home/user/MyDocs/capture%d/shift2",nsaves);
                mkdir(dirName, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
                for(unsigned int i = 2; i < shot.size()-3; i+=3){
                    char fname[256];
                    snprintf(fname, 255, "%s/trinca%4.4d.hdrgen",dirName,i);
                    FILE* hdrgen = fopen(fname,"wt");
                    char trinca[1024];
                    // Gain as a float from 1.0 to 32 (equivalent to ISO 100-3200)
                    sprintf(trinca, "foto%4.4d.jpg %f 2.8 %d 0\nfoto%4.4d.jpg %f 2.8 %d 0\nfoto%4.4d.jpg %f 2.8 %d 0",
                                    i,1.0/((float)shot[i].exposure /1000000.0),(int)(shot[i].gain*100.f + 0.5),
                                    i+1,1./((float)shot[i+1].exposure /1000000.),(int)(shot[i+1].gain*100.f + 0.5),
                                    i+2,1./((float)shot[i+2].exposure /1000000.), (int)(shot[i+2].gain*100.f + 0.5) );
                    fprintf(hdrgen,trinca);
                    fclose(hdrgen);
                }
                emit finishSave();
                sprintf(dirName,"/home/user/MyDocs/capture%d",nsaves+1);
                mkdir(dirName, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);

                nsave = fopen("/home/user/MyDocs/nsave.conf","wt");
                fprintf(nsave,"%d",nsaves+1);
                fclose(nsave);
                path = dirName;
                nbursts = aux = 0;

               sensor.stream(viewfinder);
            }
        }
        else //just streams the viewfinder
        {
            FCam::Frame f = sensor.getFrame();
            autoFocus.update(f);
            FCam::autoExpose(&viewfinder, f);
            autoWhiteBalance(&viewfinder, f);
            sensor.stream(viewfinder);
        }
    }
    sensor.stopStreaming();
    while (sensor.shotsPending()) sensor.getFrame(); //consumes the streaming queue
}

